/**************************************************************************   
 *                                                                        *
 *    File Name : lu.c                                                    *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for LU decomposition   *
 *                for VDCE application.                                   *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores L and U matrix to      *
 *                buffers so that they can be sent by vdce_send.          *
 *                                                                        *
 **************************************************************************/




#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "vdce.h"

#define inter_comm "??"

extern	char	*vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, outch1, outch2;
  char	**buf;
  char tmp[10];
  char *temp;
  int i, j, k;
  int matrix_size, type, size;
  int to, row, load, from;
  float **Matrix, **submatrix, **L, **U;
  float element, zero;
  char	*out1_msg;
  char	*out2_msg;

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  file_name[100];
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf1;
  char  *send_buf2;

  timer_reset();

  task_name = (char *) malloc(100*sizeof(char ));
  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();
  sprintf(file_name, "/tmp/%d_%s", task_id, task_name);
 
  fd = fopen(file_name, "w");

  inch1 = atoi(argv[1]);
  outch1 = atoi(argv[2]);
  outch2 = atoi(argv[3]);

  fprintf(fd, "in_port1 = %d, out_port1 = %d , out_port2 = %d\n",
	  inch1,outch1,outch2);

  fflush(fd);

  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }

  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    fprintf(fd,"vdce receive\n"); 
    fflush(fd);

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

      if (nsize <= 0 ) {
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */

    fprintf(fd,"external file\n"); 
    fflush(fd);

    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );
    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

  compute_lu(Matrix, matrix_size);

  L =  (float **) malloc(matrix_size * sizeof(float *));
  if (L == NULL) {
    fprintf(fd, "malloc error at L");
  }
  for ( i = 0 ; i < matrix_size ; i++ ) {
    L[i] = (float *) malloc(matrix_size * sizeof(float));
    if (L[i] == NULL) {
      fprintf(fd, "malloc error at L[%d]", i);
    }
  }

  U =  (float **) malloc(matrix_size * sizeof(float *));
  if (U == NULL) {
    fprintf(fd, "malloc error at U");
  }
  for ( i = 0 ; i < matrix_size ; i++ ) {
    U[i] = (float *) malloc(matrix_size * sizeof(float));
    if (U[i] == NULL) {
      fprintf(fd, "malloc error at U[%d]", i);
    }
  }

  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( i > j ) {
	L[i][j] = Matrix[i][j];
	U[i][j] = 0.;
      }
      else if ( i < j ) {
	L[i][j] = 0.;
	U[i][j] = Matrix[i][j];
      }
      else {
	L[i][j] = 1.;
	U[i][j] = Matrix[i][j];
      }
    }
  }

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
    fprintf(fd,"vdce send L\n"); 
    fflush(fd);

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      /* write L part */
      send_buf1 = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf1 == NULL) {
        fprintf(fd, "malloc error at send_buf1");
      }
      total_size = sizeof(send_buf1);
      bzero(tmp, 10);
      bzero(send_buf1, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf1, tmp);
      element = 1.0;
      zero = 0.0;
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", L[i][j]);
	  strcat(send_buf1, tmp);
	}
      }
      strcat(send_buf1,"\n");

      fprintf(fd, "L :\n \n" );

    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);

      send_buf1 = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf1, sizeof(matrix_size));

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(L[i], send_buf1 + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }

    }

    vdce_send(outch1, send_buf1, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", L[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch2);
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
    fprintf(fd,"vdce send U\n"); 
    fflush(fd);

    if ( vdce_data_represent(outch2, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      /* write U part */
      send_buf2 = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf2 == NULL) {
        fprintf(fd, "malloc error at send_buf2");
      }
      total_size = sizeof(send_buf2);
      bzero(tmp, 10);
      bzero(send_buf2, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf2, tmp);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", U[i][j]);
	  strcat(send_buf2, tmp);
	}
      }
      strcat(send_buf2,"\n");

      fprintf(fd, "U :\n \n" );

    }
    else {
      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);

      send_buf2 = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf2, sizeof(matrix_size));

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(U[i], send_buf2 + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }
    }

    vdce_send(outch2, send_buf2, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", U[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

  total_elapsed = timer_elapsed();
	
  fprintf(fd,"%s%d  %d\n", task_name, task_id, total_elapsed);
  fclose(fd);

  report_execution_time(2,total_elapsed);
}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"LU");
	
  vdce_set_inputs(mytaskinfo,"A");

  vdce_set_outputs(mytaskinfo,"L");
  vdce_set_outputs(mytaskinfo,"U");
}

compute_lu(Matrix, matrix_size)
float **Matrix;
int   matrix_size;
{

  int i, j, k;
  int current_id, current_row;
  int int_size, float_size, size;
  int current_slave;
  float Base, sum;
  
  /* k is the actual row number that is changed..
     current_slave is a slave number in which the row is changed.
     current_row is the row that is changed in each slave. */

  for ( j = 1 ; j < matrix_size ; j++ ) {
    Matrix[j][0] = Matrix[j][0] / Matrix[0][j];
  }

  for ( i = 1 ; i < matrix_size ; i++ ) {
    for ( k = 0 ; k < i - 1 ; k++ ) {
      Matrix[i][i] = Matrix[i][i] - Matrix[i][k] * Matrix[k][i];
    }

    for ( j = i + 1 ; j < matrix_size ; j++ ) {

      for ( k = 0 ; k < i - 1 ; k++ ) {
	Matrix[i][j] = Matrix[i][j] - Matrix[i][k] * Matrix[k][j];
      }

      for ( k = 0 ; k < i - 1 ; k++ ) {
	Matrix[j][i] = Matrix[j][i] - Matrix[j][k] * Matrix[k][i];
      }

      Matrix[j][i] = Matrix[j][i] / Matrix[i][i];
    }
  }
}
